#  Copyright (c) 2024 Mira Geoscience Ltd.
#
#  This file is part of geoh5py.
#
#  geoh5py is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  geoh5py is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=too-many-ancestors

from __future__ import annotations

import logging
import uuid
from abc import ABC, abstractmethod
from typing import cast

import numpy as np

from ...data import Data, ReferencedData
from ..curve import Curve
from ..object_type import ObjectType

logger = logging.getLogger(__name__)


class BaseElectrode(Curve, ABC):
    _potential_electrodes: PotentialElectrode | None = None
    _current_electrodes: CurrentElectrode | None = None

    def __init__(self, object_type: ObjectType, **kwargs):
        self._ab_cell_id: ReferencedData | None = None

        super().__init__(object_type, **kwargs)

    @property
    def ab_cell_id(self) -> ReferencedData | None:
        """
        Reference data entity mapping cells to a unique current dipole.
        """
        if getattr(self, "_ab_cell_id", None) is None:
            child = self.get_data("A-B Cell ID")
            if any(child) and isinstance(child[0], ReferencedData):
                self.ab_cell_id = child[0]

        if getattr(self, "_ab_cell_id", None) is not None:
            return self._ab_cell_id

        return None

    @ab_cell_id.setter
    def ab_cell_id(self, data: Data | np.ndarray):
        if isinstance(data, Data):
            if not isinstance(data, ReferencedData):
                raise TypeError(f"ab_cell_id must be of type {ReferencedData}")

            if data.parent.uid == self.uid:
                self._ab_cell_id = data
            else:
                self._ab_cell_id = cast(ReferencedData, data.copy(parent=self))
        else:
            if data.dtype != np.int32:
                logger.info("ab_cell_id values will be converted to type 'int32'")

            if any(self.get_data("A-B Cell ID")):
                child = self.get_data("A-B Cell ID")[0]
                if isinstance(child, ReferencedData):
                    child.values = data.astype(np.int32)
            else:
                complement: CurrentElectrode | PotentialElectrode = (
                    self.current_electrodes
                    if isinstance(self, PotentialElectrode)
                    else self.potential_electrodes
                )

                if complement is not None and complement.ab_cell_id is not None:
                    entity_type = complement.ab_cell_id.entity_type
                else:
                    value_map = {ii: str(ii) for ii in range(data.max() + 1)}
                    value_map[0] = "Unknown"
                    entity_type = {  # type: ignore
                        "primitive_type": "REFERENCED",
                        "value_map": value_map,
                    }

                data = self.add_data(
                    {
                        "A-B Cell ID": {
                            "values": data.astype(np.int32),
                            "association": "CELL",
                            "entity_type": entity_type,
                        }
                    }
                )

                if isinstance(data, ReferencedData):
                    self._ab_cell_id = data

    @property
    def ab_map(self) -> dict | None:
        """
        Get the ReferenceData.value_map of the ab_value_id
        """
        if isinstance(self.ab_cell_id, ReferencedData):
            return self.ab_cell_id.value_map
        return None

    def copy(
        self,
        parent=None,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        cell_mask: np.ndarray | None = None,
        **kwargs,
    ):
        """
        Sub-class extension of :func:`~geoh5py.objects.cell_object.CellObject.copy`.
        """
        if parent is None:
            parent = self.parent

        omit_list = [
            "_ab_cell_id",
            "_metadata",
            "_potential_electrodes",
            "_current_electrodes",
        ]
        new_entity = super().copy(
            parent=parent,
            clear_cache=clear_cache,
            copy_children=copy_children,
            mask=mask,
            cell_mask=cell_mask,
            omit_list=omit_list,
            **kwargs,
        )

        if self.cells is not None:
            if mask is not None:
                cell_mask = np.all(mask[self.cells], axis=1)
            else:
                cell_mask = np.ones(self.cells.shape[0], dtype=bool)

            if self.ab_cell_id is not None and self.ab_cell_id.values is not None:
                new_entity.ab_cell_id = self.ab_cell_id.values[cell_mask]

        complement: CurrentElectrode | PotentialElectrode = (
            self.current_electrodes
            if isinstance(self, PotentialElectrode)
            else self.potential_electrodes
        )

        # Set the mask of the complement
        if (
            new_entity.ab_cell_id is not None
            and complement is not None
            and complement.ab_cell_id is not None
            and complement.ab_cell_id.values is not None
            and complement.vertices is not None
            and complement.cells is not None
        ):
            intersect = np.intersect1d(
                new_entity.ab_cell_id.values,
                complement.ab_cell_id.values,
            )
            cell_mask = np.r_[
                [(val in intersect) for val in complement.ab_cell_id.values]
            ]

            # Convert cell indices to vertex indices
            mask = np.zeros(complement.vertices.shape[0], dtype=bool)
            mask[complement.cells[cell_mask, :]] = True

            new_complement = super(Curve, complement).copy(  # type: ignore
                parent=parent,
                omit_list=omit_list,
                copy_children=copy_children,
                clear_cache=clear_cache,
                mask=mask,
                cell_mask=cell_mask,
            )

            if isinstance(self, PotentialElectrode):
                new_entity.current_electrodes = new_complement
            else:
                new_entity.potential_electrodes = new_complement

            if new_complement.ab_cell_id is None and complement.ab_cell_id is not None:
                new_complement.ab_cell_id = complement.ab_cell_id.values[cell_mask]

            # Re-number the ab_cell_id
            value_map = {
                val: ind
                for ind, val in enumerate(
                    np.r_[0, np.unique(new_entity.current_electrodes.ab_cell_id.values)]
                )
            }
            new_map = {
                val: new_entity.current_electrodes.ab_cell_id.value_map.map[val]
                for val in value_map.values()
            }
            new_complement.ab_cell_id.values = np.asarray(
                [value_map[val] for val in new_complement.ab_cell_id.values]
            )
            new_entity.ab_cell_id.values = np.asarray(
                [value_map[val] for val in new_entity.ab_cell_id.values]
            )
            new_entity.ab_cell_id.value_map.map = new_map

        return new_entity

    @property
    @abstractmethod
    def current_electrodes(self):
        """
        The associated current_electrodes (transmitters)
        """

    @classmethod
    @abstractmethod
    def default_type_uid(cls) -> uuid.UUID:
        """Default unique identifier. Implemented on the child class."""

    @Curve.metadata.setter  # type: ignore
    def metadata(self, values: dict | None):
        if isinstance(values, dict):
            default_keys = ["Current Electrodes", "Potential Electrodes"]

            if self.metadata:
                existing_keys = self.metadata.copy()
                existing_keys.update(values)
            else:
                existing_keys = values

            # check if metadata has the required keys
            if not all(key in existing_keys for key in default_keys):
                raise ValueError(f"Input metadata must have for keys {default_keys}")

            for key in default_keys:
                if self.workspace.get_entity(existing_keys[key])[0] is None:
                    raise KeyError(f"Input {key} uuid not present in Workspace")

        super(Curve, Curve).metadata.fset(self, values)  # type: ignore

    @property
    @abstractmethod
    def potential_electrodes(self):
        """
        The associated potential_electrodes (receivers)
        """


class PotentialElectrode(BaseElectrode):
    """
    Ground potential electrode (receiver).
    """

    __TYPE_UID = uuid.UUID("{275ecee9-9c24-4378-bf94-65f3c5fbe163}")

    @property
    def current_electrodes(self):
        """
        The associated current electrode object (sources).
        """
        if getattr(self, "_current_electrodes", None) is None:
            if self.metadata is not None and "Current Electrodes" in self.metadata:
                transmitter = self.metadata["Current Electrodes"]
                transmitter_entity = self.workspace.get_entity(transmitter)[0]

                if isinstance(transmitter_entity, CurrentElectrode):
                    self._current_electrodes = transmitter_entity

        return self._current_electrodes

    @current_electrodes.setter
    def current_electrodes(self, current_electrodes: CurrentElectrode):
        if not isinstance(current_electrodes, CurrentElectrode):
            raise TypeError(
                f"Provided current_electrodes must be of type {CurrentElectrode}. "
                f"{type(current_electrodes)} provided."
            )

        metadata = {
            "Current Electrodes": current_electrodes.uid,
            "Potential Electrodes": self.uid,
        }

        self.metadata = metadata
        current_electrodes.metadata = metadata

        if isinstance(current_electrodes.ab_cell_id, ReferencedData) and isinstance(
            self.ab_cell_id, ReferencedData
        ):
            self.ab_cell_id.entity_type = current_electrodes.ab_cell_id.entity_type

    @property
    def potential_electrodes(self):
        """
        The associated potential_electrodes (receivers)
        """
        return self

    @classmethod
    def default_type_uid(cls) -> uuid.UUID:
        """
        :return: Default unique identifier
        """
        return cls.__TYPE_UID


class CurrentElectrode(BaseElectrode):
    """
    Ground direct current electrode (transmitter).
    """

    __TYPE_UID = uuid.UUID("{9b08bb5a-300c-48fe-9007-d206f971ea92}")

    def __init__(self, object_type: ObjectType, **kwargs):
        self._current_line_id: uuid.UUID | None = None

        super().__init__(object_type, **kwargs)

    @classmethod
    def default_type_uid(cls) -> uuid.UUID:
        """
        :return: Default unique identifier
        """
        return cls.__TYPE_UID

    @property
    def current_electrodes(self):
        """
        The associated current electrode object (sources).
        """
        return self

    @current_electrodes.setter
    def current_electrodes(self, _):
        """"""

    @property
    def potential_electrodes(self) -> PotentialElectrode | None:
        """
        The associated potential_electrodes (receivers)
        """
        if getattr(self, "_potential_electrodes", None) is None:
            if self.metadata is not None and "Potential Electrodes" in self.metadata:
                potential = self.metadata["Potential Electrodes"]
                potential_entity = self.workspace.get_entity(potential)[0]

                if isinstance(potential_entity, PotentialElectrode):
                    self._potential_electrodes = potential_entity

        return self._potential_electrodes

    @potential_electrodes.setter
    def potential_electrodes(self, potential_electrodes: PotentialElectrode):
        if not isinstance(potential_electrodes, PotentialElectrode):
            raise TypeError(
                f"Provided potential_electrodes must be of type {PotentialElectrode}. "
                f"{type(potential_electrodes)} provided."
            )

        metadata = {
            "Current Electrodes": self.uid,
            "Potential Electrodes": potential_electrodes.uid,
        }

        self.metadata = metadata
        potential_electrodes.metadata = metadata

        if isinstance(potential_electrodes.ab_cell_id, ReferencedData) and isinstance(
            self.ab_cell_id, ReferencedData
        ):
            potential_electrodes.ab_cell_id.entity_type = self.ab_cell_id.entity_type

    def add_default_ab_cell_id(self):
        """
        Utility function to set ab_cell_id's based on curve cells.
        """
        if getattr(self, "cells", None) is None or self.n_cells is None:
            raise AttributeError(
                "Cells must be set before assigning default ab_cell_id"
            )

        data = np.arange(self.n_cells) + 1
        value_map = {ii: str(ii) for ii in range(self.n_cells + 1)}
        value_map[0] = "Unknown"
        ab_cell_id = self.add_data(
            {
                "A-B Cell ID": {
                    "values": data,
                    "association": "CELL",
                    "entity_type": {
                        "primitive_type": "REFERENCED",
                        "value_map": value_map,
                    },
                }
            }
        )
        if isinstance(ab_cell_id, ReferencedData):
            ab_cell_id.entity_type.name = "A-B"
            self._ab_cell_id = ab_cell_id



#  Copyright (c) 2024 Mira Geoscience Ltd.
#
#  This file is part of geoh5py.
#
#  geoh5py is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  geoh5py is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Literal, cast, get_args

import numpy as np

from ..shared import EntityType
from .color_map import ColorMap
from .geometric_data_constants import GeometricDataConstants
from .primitive_type_enum import PrimitiveTypeEnum
from .reference_value_map import ReferenceValueMap

if TYPE_CHECKING:
    from ..workspace import Workspace
    from .data import Data  # noqa: F401


ColorMapping = Literal["linear", "equal_area", "logarithmic", "cdf", "missing"]


class DataType(EntityType):
    # pylint: disable=too-many-arguments
    """
    DataType class.

    Controls all the attributes of the data type for displays in Geoscience ANALYST.

    :param workspace: An active Workspace.
    :param primitive_type: The primitive type of the data.
    :param color_map: The colormap used for plotting.
    :param hidden: If the data are hidden or not.
    :param mapping: The type of color stretching to plot the colormap.
    :param number_of_bins: The number of bins used by the histogram.
    :param transparent_no_data: If the no data values are displayed as transparent or not.
    :param units: The type of the units of the data.
    :param value_map: Reference value map for to map index with description.
    :param kwargs: Additional keyword arguments to set as attributes
        (see :obj:`...shared.entity_type.EntityType`).
    """

    _attribute_map = EntityType._attribute_map.copy()
    _attribute_map.update(
        {
            "Hidden": "hidden",
            "Mapping": "mapping",
            "Number of bins": "number_of_bins",
            "Primitive type": "primitive_type",
            "Transparent no data": "transparent_no_data",
        }
    )

    def __init__(
        self,
        workspace: Workspace,
        primitive_type: type[Data] | PrimitiveTypeEnum | str | None = None,
        color_map: ColorMap | None = None,
        hidden: bool = False,
        mapping: ColorMapping = "equal_area",
        number_of_bins: int | None = None,
        transparent_no_data: bool = True,
        units: str | None = None,
        value_map: dict[int, str] | ReferenceValueMap | None = None,
        **kwargs,
    ):
        super().__init__(workspace, **kwargs)
        self.color_map = color_map
        self.hidden = hidden
        self.mapping = mapping
        self.number_of_bins = number_of_bins
        self.primitive_type = primitive_type  # type: ignore
        self.transparent_no_data = transparent_no_data
        self.units = units
        self.value_map = value_map  # type: ignore

    @classmethod
    def _for_geometric_data(cls, workspace: Workspace, uid: uuid.UUID) -> DataType:
        """
        Get the data type for geometric data.

        :param workspace: An active Workspace.
        :param uid: The uid of the existing data type to get.

        :return: A new instance of DataType.
        """
        geom_primitive_type = GeometricDataConstants.primitive_type()
        data_type = cast(DataType, workspace.find_type(uid, DataType))
        if data_type is not None:
            if not data_type.primitive_type == geom_primitive_type:
                raise ValueError(
                    f"Data type with uid {uid} is not of type {geom_primitive_type}"
                )
            return data_type
        return cls(workspace, uid=uid, primitive_type=geom_primitive_type)

    @property
    def color_map(self) -> ColorMap | None:
        r"""
        The Colormap used for plotting

        The colormap can be set from a dictionary of sorted values with
        corresponding RGBA color.
        Or from a numpy array containing the RGBA values.

        .. code-block:: python

            color_map = {
                val_1: [r_1, g_1, b_1, a_1],
                ...,
                val_i: [r_i, g_i, b_i, a_i]
            }

        It can be set to None if non-existing.
        """
        return self._color_map

    @color_map.setter
    def color_map(self, color_map: ColorMap | dict | np.ndarray | None):
        if isinstance(color_map, dict):
            color_map = ColorMap(**color_map)

        elif isinstance(color_map, np.ndarray):
            color_map = ColorMap(values=color_map)

        if not isinstance(color_map, (ColorMap, type(None))):
            raise TypeError(
                f"Input value for 'color_map' must be of type {ColorMap},"
                f"numpy.ndarray or dict with 'values'."
            )

        if isinstance(color_map, ColorMap):
            color_map.parent = self

        self._color_map: ColorMap | None = color_map

        self.workspace.update_attribute(self, "color_map")

    @classmethod
    def for_x_data(cls, workspace: Workspace) -> DataType:
        """
        Get the data type for x data.

        :param workspace: An active Workspace.

        :return: A new instance of DataType.
        """
        return cls._for_geometric_data(
            workspace, GeometricDataConstants.x_datatype_uid()
        )

    @classmethod
    def for_y_data(cls, workspace: Workspace) -> DataType:
        """
        Get the data type for y data.

        :param workspace: An active Workspace.

        :return: A new instance of DataType.
        """
        return cls._for_geometric_data(
            workspace, GeometricDataConstants.y_datatype_uid()
        )

    @classmethod
    def for_z_data(cls, workspace: Workspace) -> DataType:
        """
        Get the data type for z data.

        :param workspace: An active Workspace.

        :return: A new instance of DataType.
        """
        return cls._for_geometric_data(
            workspace, GeometricDataConstants.z_datatype_uid()
        )

    @property
    def hidden(self) -> bool:
        """
        If the data are hidden or not.
        """
        return self._hidden

    @hidden.setter
    def hidden(self, value: bool):
        if not isinstance(value, bool) and value != 1 and value != 0:
            raise TypeError(f"hidden must be a bool, not {type(value)}")

        self._hidden: bool = bool(value)

        self.workspace.update_attribute(self, "attributes")

    @property
    def mapping(self) -> str:
        """
        The type of color stretching to plot the colormap.
        It chan be one of the following:
        'linear', 'equal_area', 'logarithmic', 'cdf', 'missing'
        """
        return self._mapping

    @mapping.setter
    def mapping(self, value: ColorMapping):
        if value not in get_args(ColorMapping):
            raise ValueError(
                f"Mapping {value} was provided but should be one of {get_args(ColorMapping)}"
            )
        self._mapping: str = value

        self.workspace.update_attribute(self, "attributes")

    @property
    def number_of_bins(self) -> int | None:
        """
        The number of bins used by the histogram.
        It can be None if no histogram is used.
        """
        return self._number_of_bins

    @number_of_bins.setter
    def number_of_bins(self, n_bins: int | None):
        if n_bins is None:
            pass
        elif not isinstance(n_bins, (int, np.integer)) or n_bins < 1:
            raise ValueError(
                f"Number of bins should be an integer greater than 0 or None, not {n_bins}"
            )

        self._number_of_bins: int | None = n_bins

        self.workspace.update_attribute(self, "attributes")

    @property
    def primitive_type(self) -> PrimitiveTypeEnum | None:
        """
        The primitive type of the data.
        """
        return self._primitive_type

    @primitive_type.setter
    def primitive_type(self, value: str | type[Data] | PrimitiveTypeEnum | None):
        if isinstance(value, str):
            value = getattr(PrimitiveTypeEnum, value.replace("-", "_").upper())
        elif hasattr(value, "primitive_type"):
            value = getattr(value, "primitive_type")()
        if not isinstance(value, (PrimitiveTypeEnum, type(None))):
            raise ValueError(
                f"Primitive type value must be of type {PrimitiveTypeEnum}, find {type(value)}"
            )

        self._primitive_type = value

    @property
    def transparent_no_data(self) -> bool:
        """
        If the no data values are displayed as transparent or not.
        """
        return self._transparent_no_data

    @transparent_no_data.setter
    def transparent_no_data(self, value: bool):
        if not isinstance(value, bool) and value != 1 and value != 0:
            raise TypeError(f"transparent_no_data must be a bool, not {type(value)}")
        self._transparent_no_data = bool(value)

        self.workspace.update_attribute(self, "attributes")

    @property
    def units(self) -> str | None:
        """
        The type of the units of the data.
        """
        return self._units

    @units.setter
    def units(self, unit: str | None):
        if not isinstance(unit, (str, type(None))):
            raise TypeError(f"units must be a string, not {type(unit)}")
        self._units = unit

        self.workspace.update_attribute(self, "attributes")

    @staticmethod
    def validate_data_type(workspace: Workspace, attribute_dict: dict):
        """
        Get a dictionary of attributes and validate the type of data.

        :param workspace: An active Workspace.
        :param attribute_dict: A dictionary of attributes of the new Datatype to create.

        :return: A new instance of DataType.
        """

        entity_type = attribute_dict.get("entity_type")
        if entity_type is None:
            primitive_type = attribute_dict.get("type")
            if primitive_type is not None:
                assert (
                    primitive_type.upper() in PrimitiveTypeEnum.__members__
                ), f"Data 'type' should be one of {PrimitiveTypeEnum.__members__}"
                entity_type = {"primitive_type": primitive_type.upper()}
            else:
                values = attribute_dict.get("values")
                if values is None or (
                    isinstance(values, np.ndarray)
                    and np.issubdtype(values.dtype, np.floating)
                ):
                    entity_type = {"primitive_type": "FLOAT"}

                elif isinstance(values, np.ndarray) and (
                    np.issubdtype(values.dtype, np.integer)
                ):

                    entity_type = {"primitive_type": "INTEGER"}
                elif isinstance(values, str) or (
                    isinstance(values, np.ndarray) and values.dtype.kind in ["U", "S"]
                ):
                    entity_type = {"primitive_type": "TEXT"}
                elif isinstance(values, np.ndarray) and (values.dtype == bool):
                    entity_type = {"primitive_type": "BOOLEAN"}
                else:
                    raise NotImplementedError(
                        "Only add_data values of type FLOAT, INTEGER,"
                        "BOOLEAN and TEXT have been implemented"
                    )
        elif isinstance(entity_type, EntityType) and (
            (entity_type.uid not in getattr(workspace, "_types"))
            or (entity_type.workspace != workspace)
        ):
            return entity_type.copy(workspace=workspace)

        return entity_type

    @property
    def value_map(self) -> ReferenceValueMap | None:
        r"""
        Reference value map for to map index with description.

        The value_map can be set from a dictionary of sorted values int
        values with text description.

        .. code-block:: python

            value_map = {
                val_1: str_1,
                ...,
                val_i: str_i
            }

        """
        return self._value_map

    @value_map.setter
    def value_map(self, value_map: dict[int, str] | ReferenceValueMap | None):
        if isinstance(value_map, dict):
            value_map = ReferenceValueMap(value_map)
        if not isinstance(value_map, (ReferenceValueMap, type(None))):
            raise TypeError(
                f"'value_map' must be a {dict} or {ReferenceValueMap} or {type(None)}."
            )

        self._value_map: ReferenceValueMap | None = value_map

        self.workspace.update_attribute(self, "value_map")



#  Copyright (c) 2024 Mira Geoscience Ltd.
#
#  This file is part of geoh5py.
#
#  geoh5py is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  geoh5py is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from geoh5py.ui_json.forms import FormParameter
    from geoh5py.ui_json.parameters import Parameter


# pylint: disable=too-few-public-methods


class ValueAccess:
    """
    Descriptor to elevate underlying member values within 'FormParameter'.

    :param private: Name of private attribute.
    """

    def __init__(self, private: str):
        self.private: str = private

    def __get__(self, obj: Parameter, objtype=None):
        return getattr(obj, self.private).value

    def __set__(self, obj: Parameter, value):
        setattr(getattr(obj, self.private), "value", value)


class FormValueAccess:
    """
    Descriptor to elevate underlying member values within 'FormParameter'.

    :param private: Name of private attribute.
    """

    def __init__(self, private: str):
        self.private: str = private

    def __get__(self, obj: FormParameter, objtype=None):
        return getattr(obj, self.private).value

    def __set__(self, obj: FormParameter, value):
        setattr(getattr(obj, self.private), "value", value)
        obj._active_members.append(self.private[1:])



#  Copyright (c) 2024 Mira Geoscience Ltd.
#
#  This file is part of geoh5py.
#
#  geoh5py is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  geoh5py is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Any

from geoh5py.data import (
    BooleanData,
    DatetimeData,
    FloatData,
    IntegerData,
    ReferencedData,
    TextData,
)
from geoh5py.shared.exceptions import AggregateValidationError, BaseValidationError
from geoh5py.shared.utils import SetDict
from geoh5py.ui_json.descriptors import FormValueAccess
from geoh5py.ui_json.enforcers import EnforcerPool
from geoh5py.ui_json.parameters import (
    BoolParameter,
    FloatParameter,
    IntegerParameter,
    NumericParameter,
    Parameter,
    StringListParameter,
    StringParameter,
    TypeRestrictedParameter,
    TypeUIDRestrictedParameter,
    ValueRestrictedParameter,
)


class MemberKeys:
    """Converts in and out of camel (ui.json) and snake (python) case"""

    camel_to_snake: dict[str, str] = {
        "groupOptional": "group_optional",
        "dependencyType": "dependency_type",
        "groupDependency": "group_dependency",
        "groupDependencyType": "group_dependency_type",
        "lineEdit": "line_edit",
        "choiceList": "choice_list",
        "fileDescription": "file_description",
        "fileType": "file_type",
        "fileMulti": "file_multi",
        "meshType": "mesh_type",
        "dataType": "data_type",
        "dataGroupType": "data_group_type",
        "isValue": "is_value",
    }

    @property
    def snake_to_camel(self) -> dict[str, str]:
        """Gives the inverse map to camel_to_snake."""
        return {v: k for k, v in self.camel_to_snake.items()}

    def map_key(self, key: str, convention: str = "snake"):
        """Map a string from snake to camel or vice versa."""

        if convention == "snake":
            out = self.camel_to_snake.get(key, key)
        elif convention == "camel":
            out = self.snake_to_camel.get(key, key)
        else:
            raise ValueError("Convention must be 'snake' or 'camel'.")

        return out

    def map(self, collection: dict[str, Any], convention="snake"):
        """Map a dictionary from snake to camel or vice versa."""
        return {self.map_key(k, convention): v for k, v in collection.items()}


MEMBER_KEYS = MemberKeys()


class FormParameter:  # pylint: disable=too-many-instance-attributes
    """
    Base class for parameters that create visual ui elements from a form.

    :param name: Parameter name.
    :param label: Label for ui element.
    :param value: The parameter's value.
    :param enabled: If False, ui element is rendered grey and value is
        treated as None.
    :param optional: If True, ui element is rendered with a checkbox to
        control the enabled state.
    :param group_optional: If True, ui group is rendered with a checkbox
        that control the enabled state of all group members.
    :param main: Controls whether ui element will render in the general
        parameters tab (True) or optional parameters (False).
    :param group: Grouped ui elements will be rendered within a named
        box.
    :param dependency: Name of parameter that controls the enabled or
        visible state of the ui element.
    :param dependency_type: Controls whether the ui element is enabled
        or visible when the dependency is enabled if optional or True
        if a bool type.
    :param group_dependency: Name of parameter that controls the enabled
        or visible state of the ui group.
    :param group_dependency_type: Controls whether the ui group is
        enabled or visible when the group dependency is enabled if
        optional or True if a bool type.
    :param tooltip: String rendered on hover over ui element.

    :note: Standardized form members are accessible through public namespace
        by way of the FormValueAccess descriptor.
    """

    static_validations = {"required_form_members": ["label", "value"]}
    identifier_members: list[str] = []

    def __init__(
        self,
        name: str,
        value: Any | None = None,
        **kwargs,
    ):
        self.name: str = name
        self._value: Parameter = self._set_value_parameter(value)
        self._label = StringParameter("label")
        self._enabled = BoolParameter("enabled", value=True)
        self._optional = BoolParameter("optional")
        self._group_optional = BoolParameter("group_optional")
        self._main = BoolParameter("main", value=True)
        self._group = StringParameter("group")
        self._dependency = StringParameter("dependency")
        self._dependency_type = ValueRestrictedParameter(
            "dependency_type", ["enabled", "disabled"], value="enabled"
        )
        self._group_dependency = StringParameter("group_dependency")
        self._group_dependency_type = ValueRestrictedParameter(
            "group_dependency_type", ["enabled", "disabled"], value="enabled"
        )
        self._tooltip = StringParameter("tooltip")
        self._allow_values_access()
        self._extra_members: dict[str, Any] = {}
        self._active_members: list[str] = []
        if kwargs:
            self.register(kwargs)
        self._validations = SetDict()
        self.enforcers: EnforcerPool = EnforcerPool.from_validations(
            self.name, self.validations
        )

    @property
    def validations(self):
        """Returns a dictionary of static and inferred validations."""
        if not self._validations:
            self._validations.update(self.dynamic_validations)
            self._validations.update(self.static_validations)

        return self._validations

    @property
    def dynamic_validations(self):
        """Infer validations from parameters."""
        validations = SetDict()
        if "group_optional" in self.active:
            validations.update({"required": "group"})

        return validations

    @property
    def uijson_validations(self):
        """Validations for UIJson level enforcers."""
        validations = SetDict()
        if "dependency" in self.active:
            validations.update({"required": self.dependency})  # type: ignore

        return validations

    def form(self, use_camel: bool = False) -> dict[str, Any]:
        """
        Returns dictionary of active form members and their values.

        :param use_camel: If True, keys are converted to camel case.
        """
        form = {}
        for member in self.active:
            if member in self._extra_members:
                form[member] = self._extra_members[member]
            else:
                form[member] = getattr(self, f"_{member}").value

        if use_camel:
            form = MEMBER_KEYS.map(form, "camel")

        return form

    def register(self, members: dict[str, Any]):
        """
        Update form members with incoming values.

        :param members: Dictionary of form members and associated data.
        """

        error_list = []
        members = MEMBER_KEYS.map(members)
        for member in list(members):
            if member in self.valid_members:
                try:
                    setattr(self, member, members.pop(member))
                except BaseValidationError as err:
                    error_list.append(err)

        if error_list:
            if len(error_list) == 1:
                raise error_list.pop()
            raise AggregateValidationError(self.name, error_list)

        self._extra_members.update(members)

    def validate(self):
        """Validates form data against the pool of enforcers."""
        self.enforcers.enforce(self.form())

    @property
    def valid_members(self) -> list[str]:
        """Recognized form member names."""
        exclusions = ["_extra_members", "_active_members", "_validations"]
        private_attrs = [k for k in self.__dict__ if k.startswith("_")]
        return [k[1:] for k in private_attrs if k not in exclusions]

    @property
    def active(self) -> list[str]:
        """
        Returns names of active form members.

        :return: List of active form members.  These will include any members
            that were:
                1. Provided during construction.
                2. Updated through the 'register' method.
                3. Updated through member setters.
        """
        active = ["value"] + self._active_members + list(self._extra_members)
        return list(dict.fromkeys(active))  # Unique while preserving order

    @classmethod
    def is_form(cls, form: dict[str, Any]) -> bool:
        """Returns True if form contains any identifier members."""
        id_members = cls.identifier_members
        form_members = MEMBER_KEYS.map(form)
        return any(k in form_members for k in id_members)

    @property
    def value(self):
        return self._value.value

    @value.setter
    def value(self, val):
        self._value.value = val

    def _set_value_parameter(self, value) -> Parameter:
        """Handles value argument as either a Parameter or a value."""

        if isinstance(value, Parameter):
            value.name = "value"
            out = value
        else:
            out = Parameter("value", value)

        return out

    def _allow_values_access(self):
        """Valid members public attr accesses underlying parameter value."""
        for member in self.valid_members:
            if member not in dir(self):
                setattr(self.__class__, member, FormValueAccess(f"_{member}"))

    def __str__(self):
        return f"<{type(self).__name__}> : '{self.name}' -> {self.value}"

    def __contains__(self, item):
        return MEMBER_KEYS.map_key(item) in self.active


class StringFormParameter(FormParameter):
    """String parameter type."""

    def __init__(self, name, value=None, **kwargs):
        value = StringParameter("value", value=value)
        super().__init__(name, value=value, **kwargs)


class BoolFormParameter(FormParameter):
    """Boolean parameter type."""

    def __init__(self, name, value: bool = False, **kwargs):
        param = BoolParameter("value", value=value)
        super().__init__(name, value=param, **kwargs)


class IntegerFormParameter(FormParameter):
    """
    Integer parameter type.

    :param min: Minimum value for ui element.
    :param max: Maximum value for ui element.
    """

    identifier_members: list[str] = []

    def __init__(self, name, value=None, **kwargs):
        self._min = IntegerParameter("min")
        self._max = IntegerParameter("max")
        value = IntegerParameter("value", value=value)
        super().__init__(name, value=value, **kwargs)


class FloatFormParameter(FormParameter):
    """
    Float parameter type.

    :param min: Minimum value for ui element.
    :param max: Maximum value for ui element.
    :param precision: Number of decimal places to display in ui element.
    :param line_edit: If False, the ui element incluces a spinbox.
    """

    identifier_members: list[str] = ["precision", "line_edit"]

    def __init__(self, name, value=None, **kwargs):
        self._min = FloatParameter("min")
        self._max = FloatParameter("max")
        self._precision = IntegerParameter("precision")
        self._line_edit = BoolParameter("line_edit", value=True)
        value = FloatParameter("value", value=value)
        super().__init__(name, value=value, **kwargs)


class ChoiceStringFormParameter(FormParameter):
    """
    Choice string parameter type.

    :param choice_list: List of choices for ui dropdown.
    """

    identifier_members: list[str] = ["choice_list"]
    static_validations = {"required_form_members": ["choice_list"]}

    def __init__(self, name, choice_list, value=None, **kwargs):
        self._choice_list = StringListParameter("choice_list")
        value = ValueRestrictedParameter("value", choice_list, value=value)
        super().__init__(name, value=value, choice_list=choice_list, **kwargs)


class FileFormParameter(FormParameter):
    """
    File parameter type.

    :param file_description: list of file descriptions for each file type.
    :param file_type: list of file extensions to filter directory on.
    :param file_multi: Allow multiple files to be selected from dropdown.
    """

    identifier_members: list[str] = ["file_description", "file_type", "file_multi"]
    static_validations = {"required_form_members": ["file_description", "file_type"]}

    def __init__(self, name, value=None, **kwargs):
        self._file_description = StringListParameter("file_description")
        self._file_type = StringListParameter("file_type")
        self._file_multi = BoolParameter("file_multi")
        value = StringParameter("value", value=value)
        super().__init__(name, value=value, **kwargs)


class ObjectFormParameter(FormParameter):
    """
    Object parameter type.

    :param mesh_type: list of object types (uid) that will be available in the
        dropdown.  Empty list will reveal all objects in geoh5.
    """

    identifier_members: list[str] = ["mesh_type"]
    static_validations = dict(
        FormParameter.static_validations, **{"required_form_members": ["mesh_type"]}
    )

    def __init__(self, name, mesh_type, value=None, **kwargs):
        self._mesh_type = StringListParameter("mesh_type", value=[])
        value = TypeUIDRestrictedParameter("value", mesh_type, value=value)
        super().__init__(name, value=value, **kwargs)

    @property
    def uijson_validations(self):
        """Validations for UIJson level enforcers."""
        validations = SetDict()
        if self.value is not None:
            validations.update({"required_workspace_object": self.name})

        return validations


DATA_TYPES = {
    "Integer": IntegerData,
    "Float": FloatData,
    "Text": TextData,
    "Referenced": ReferencedData,
    "DateTime": DatetimeData,
    "Boolean": BooleanData,
}


class DataFormParameter(FormParameter):
    """
    Data parameter type.

    :param parent: Name of parent object.
    :param association: Filters data to those living on vertices or cells.
    :param data_type: Filters data type.
    :param data_group_type: Filters data group type.
    """

    identifier_members: list[str] = ["data_group_type"]
    static_validations = {
        "required_form_members": ["parent", "association", "data_type"]
    }

    def __init__(self, name, data_type, value=None, **kwargs):
        self._parent = StringParameter("parent")
        self._association = ValueRestrictedParameter(
            "association", ["Vertex", "Cell", "Face"]
        )
        self._data_type = ValueRestrictedParameter("data_type", list(DATA_TYPES))
        self._data_group_type = ValueRestrictedParameter(
            "data_group_type", ["3D vector", "Dip direction & dip", "Strike & dip"]
        )
        value = TypeRestrictedParameter(
            "value", [DATA_TYPES.get(data_type, None)], value=value
        )
        super().__init__(name, value=value, data_type=data_type, **kwargs)

    @property
    def uijson_validations(self):
        """Validations for UIJson level enforcers."""

        validations = SetDict()
        if self.value is not None:
            validations.update(
                {
                    "required_workspace_object": self.name,
                    "required_object_data": [(self._parent.value, self.name)],
                }
            )

        return validations


class DataValueFormParameter(FormParameter):
    """
    Data value parameter type.

    :param parent: Name of parent object.
    :param association: Filters data to those living on vertices or cells.
    :param data_type: Filters data type.
    :param is_value: Gives ui element a button to switch between value box
        and dropdown of available properties.
    :param property: Name of property.
    """

    identifier_members: list[str] = ["is_value", "property"]
    static_validations = {
        "required_form_members": [
            "parent",
            "association",
            "data_type",
            "is_value",
            "property",
        ]
    }

    def __init__(self, name, data_type, value=None, **kwargs):
        self._parent = StringParameter("parent")
        self._association = ValueRestrictedParameter(
            "association", ["Vertex", "Cell", "Face"]
        )
        self._data_type = ValueRestrictedParameter("data_type", list(DATA_TYPES))
        self._is_value = BoolParameter("is_value")
        self._property = TypeRestrictedParameter(
            "property", [DATA_TYPES.get(data_type, type(None))]
        )
        value = NumericParameter("value", value=value)
        super().__init__(name, value=value, data_type=data_type, **kwargs)

    @property
    def value(self):
        """Form value is value of property when is_value is False."""
        val = self.property  # type: ignore # pylint: disable=no-member
        if self.is_value:  # type: ignore # pylint: disable=no-member
            val = self._value.value
        return val

    @value.setter
    def value(self, val):
        if isinstance(val, (int, float)):
            self._value.value = val
        else:
            self._property.value = val
